/*
    $Id: cpuspeed.cc 4 2003-11-07 23:06:53Z carl $

    Copyright 2002 - 2003

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    cpuspeed.cc

    Author:
        Carl Thompson <cet@carlthompson.net>

    Contributors:
        Enrico Tassi <gareuselesinge@libero.it>

    This program is only for computers with Linux kernels compiled with
    CPUFreq.  You must have a CPU that supports frequency and/or voltage
    scaling via CPUFreq to use this program.  Your kernel must be compiled to
    support the "userspace" CPUFreq governor and the "sysfs" interface on
    Linux 2.6 or the "proc" interface on Linux 2.4.

    I use this program on my eMachines M5305 laptop to increase battery life
    and control performance.
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/errno.h>
#include <stdarg.h>
#include <time.h>
#include <sys/utsname.h>

#ifndef NAME
#  define NAME "cpuspeed"
#endif

#define VERSION "1.1"

// Maximum number of speed steps.  Must be 2 or greater
#define MAX_SPEEDS  20

// To activate debug printings
#ifdef DEBUG
#   define errprintf(A...) fprintf(stderr,A)
#else
#   define errprintf(A...)
#endif

#define SYSFS_MIN_SPEED_FILE \
    "/sys/devices/system/cpu/cpu0/cpufreq/scaling_min_freq"

#define SYSFS_MAX_SPEED_FILE \
     "/sys/devices/system/cpu/cpu0/cpufreq/scaling_max_freq"

#define SYSFS_CURRENT_SPEED_FILE \
     "/sys/devices/system/cpu/cpu0/cpufreq/scaling_setspeed"

#define SYSFS_GOVERNOR_FILE \
     "/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor"

#define SYSFS_USERSPACE \
    "userspace"

#define PROC_MIN_SPEED_FILE \
    "/proc/sys/cpu/0/speed-min"

#define PROC_MAX_SPEED_FILE \
    "/proc/sys/cpu/0/speed-max"

#define PROC_CURRENT_SPEED_FILE \
    "/proc/sys/cpu/0/speed"

#define PROC_GOVERNOR_FILE \
    "/proc/cpufreq"

#define PROC_USERSPACE \
    "0%0%100%userspace"

char * MIN_SPEED_FILE;
char * MAX_SPEED_FILE;
char * CURRENT_SPEED_FILE;
char * GOVERNOR_FILE;
char * USERSPACE;

// defines what info we care about for each speed step
struct
{
    unsigned khz;
    //unsigned volts;
    //unsigned fsb;
} speeds[MAX_SPEEDS + 1];

// display an error message and exit the program
void
die(bool system_error, const char *fmt, ...)
{
    fprintf(stderr, "Error: ");

    va_list ap;
    // get variable argument list passed
    va_start(ap, fmt);
    // display message passed on stderr
    vfprintf(stderr, fmt, ap);
    va_end(ap);

    fprintf(stderr, "\n");
    if (system_error)
        fprintf(stderr, "Error: %s\n", strerror(errno));

    exit(1);
}

// read a line from a file
void
read_line(const char * filename, char * line, unsigned len)
{
    FILE * fp = fopen(filename, "r");
    if (!fp)
        die(true, "Could not open file for reading: %s", filename);
    if ( (!fgets(line, len, fp)) )
        die(true, "Could not read from file: %s", filename);
    fclose(fp);
}

// write a line to a file
void
write_line(const char * filename, const char *fmt, ...)
{
    FILE * fp = fopen(filename, "w+");
    if (!fp)
        die(true, "Could not open file for writing: %s", filename);
    va_list ap;
    // get variable argument list passed
    va_start(ap, fmt);
    if (vfprintf(fp, fmt, ap) < 0)
        die(true, "Could not write to file: %s", filename);
    va_end(ap);
    fclose(fp);
}

// read an integer value from a file
unsigned
read_value(const char * filename)
{
    char line[256];
    read_line(filename, line, sizeof line);
    return atoi(line);
}

// get the current CPU speed
unsigned get_speed()
{
    return read_value(CURRENT_SPEED_FILE);
}

// set the current CPU speed
void set_speed(unsigned value)
{
    errprintf("Setting speed to: %uKHz\n", value);
    write_line(CURRENT_SPEED_FILE, "%u\n", value);
    // give CPU / chipset voltage time to settle down
    usleep(10000);
}

// This code smoothly transitions the CPU speed from 'current' to 'target'
// instead of jumping directly to the new speed because some AMD Mobile
// Athlon parts seem to choke on large differentials causing kernel panics.
void set_speed(unsigned current, unsigned target)
{
    if (current == target)
        return;
    int delta = (current > target) ? -1 : 1;
    do
    {
        current += delta;
        set_speed(speeds[current].khz);
    }
    while (current != target);
}

// lowest speed step
unsigned last_step;

// get the speed steps supported by the CPU
void
get_supported_speeds()
{
    unsigned min = read_value(MIN_SPEED_FILE);
    unsigned max = read_value(MAX_SPEED_FILE);
    const unsigned step = 25000;

    // Go to max speed if we are not already there
    for (unsigned current = get_speed(); current < max; current += step)
        set_speed(current);

    // this code is a hack to get the various speed steps supported by the
    // CPU by looping from the maximum speed to the minimum speed and trying
    // to set every possible speed divisible by step!
    speeds[0].khz = max;
    unsigned current_speed = 0;
    for (unsigned current = max - step; current > min - step; current -= step)
    {
        set_speed(current);
        unsigned real = get_speed();
        if (real != speeds[current_speed].khz)
        {
            speeds[++current_speed].khz = real;
            if (current_speed + 1 == MAX_SPEEDS)
                break;
        }
#ifdef DEBUG
        usleep(150000);
#endif
    }
    speeds[current_speed + 1].khz = 0;

    // go back to maximum speed because program expects to start there
    set_speed(current_speed, 0);

    // the last step is the lowest found speed
    last_step = current_speed;

#ifdef DEBUG
    errprintf("Available speeds:\n");
    for (current_speed = 0; speeds[current_speed].khz; current_speed++)
        errprintf(" %2u: %9uKHz\n", current_speed, speeds[current_speed].khz);
#endif
}

// gets the elapsed total time and elapsed idle time since it was last called
void
get_times(unsigned long & total_elapsed, unsigned long & idle_elapsed)
{
    char what[32];
    unsigned long user_time, nice_time, system_time, idle_time, total_time;
    static unsigned long last_total_time = 0, last_idle_time = 0;

    bool found;
    char line[256];
    FILE * fp = fopen("/proc/stat", "r");
    if (!fp)
        die(true, "Could not open /proc/stat for reading!");
    while ( (found = fgets(line, sizeof line, fp)) )
    {
        sscanf(line, "%s %lu %lu %lu %lu", what, &user_time, &nice_time,
               &system_time, &idle_time);
        if (!strcmp(what, "cpu0"))
            break;
    }
    fclose(fp);

    if (!found)
        die(false, "Could not find entry for cpu0 in /proc/stat!?");

    // count nice time as idle time
    idle_time += nice_time;

    total_time = user_time + system_time + idle_time;
    total_elapsed = total_time - last_total_time;
    last_total_time = total_time;
    idle_elapsed = idle_time - last_idle_time;
    last_idle_time = idle_time;

    //errprintf("time: %lu    idle: %lu\n", total_elapsed, idle_elapsed);
}

// resets the elapsed total time and elapsed idle time counters
void reset_times()
{
    unsigned long dummy1, dummy2;
    get_times(dummy1, dummy2);
}

// are we currently dynamically scaling the CPU or at min or max?
enum Mode { DYNAMIC, MIN, MAX } mode;

// if CPU idle percentage is below this, CPU will be set to fastest speed
unsigned clock_up_idle_fast= 10;

// if CPU idle percentage is below this, CPU will be set to next higher speed
// if CPU idle percentage is above this, CPU will be set to next lower speed
unsigned idle_threshold = 25;

// if this is set (via command line) throttle down CPU to minimum if
// temperature is too high
const char * temperature_filename = 0;
unsigned max_temperature;

// if this is set (via command line) throttle down CPU to minimum if
// AC power is disconnected
const char * ac_filename = 0;

// handles the periodic check of idle and setting CPU speed
void
alarm_handler(int)
{
    static unsigned new_speed = 0;
    static unsigned current_speed = 0;

    // current state
    Mode state = mode;

    char line[256], *p;

    if (mode == DYNAMIC)
    {
        // check that we are not getting too hot
        if (temperature_filename)
        {
            read_line(temperature_filename, line, sizeof line);
            p = strpbrk(line, "0123456789");
            if (!p)
                die(false, "Could not find temperature in file: %s",
                    temperature_filename);
            if (strtol(p, 0, 10) > (long)max_temperature)
            {
                state = MIN;
                reset_times();
            }
        }

        // check to see if AC power is disconnected
        if (ac_filename)
        {
            read_line(ac_filename, line, sizeof line);
            if (strstr(line, "off-line"))
            {
                state = MIN;
                reset_times();
            }
        }
    }

    // figure out what our current speed should be
    switch(state)
    {
        case DYNAMIC:
        {
            unsigned long elapsed_time, idle_time;
            unsigned idle_percent;

            // get the elapsed and idle times since we last checked
            get_times(elapsed_time, idle_time);

            if (elapsed_time > 0)
            {
                idle_percent = idle_time * 100 / elapsed_time;

                if (idle_percent <= clock_up_idle_fast)
                    new_speed = 0;
                else if (idle_percent < idle_threshold && new_speed != 0)
                    new_speed--;
                else if (idle_percent > idle_threshold && speeds[new_speed + 1].khz != 0)
                    new_speed++;

                errprintf("Idle percent: %.2u   %4dKHz\n", idle_percent, speeds[new_speed].khz);
            }
            break;
        }

        case MIN:
            new_speed = last_step;
            break;

        case MAX:
            new_speed = 0;
            break;
    }

    // if the last set speed is not what it currently should be, set it
    if (current_speed != new_speed)
    {
        errprintf("current speed: %uKHz     new speed: %uKHz\n",
                  speeds[current_speed].khz, speeds[new_speed].khz);
        set_speed(current_speed, new_speed);
        current_speed = new_speed;
    }
}

// handles the USR1 signal (stay at maximum performance)
void
usr1_handler(int)
{
    mode = MAX;
}

// handles the USR2 signal (stay at minimum performance)
void
usr2_handler(int)
{
    mode = MIN;
}

// handles the HUP signal (dynamically scale performance)
void
hup_handler(int)
{
    reset_times();
    mode = DYNAMIC;
}

long
get_int(const char * s)
{
    char * end;
    long r = strtol(s, &end, 10);
    if (*end != '\0')
        die(false, "Not an integer: %s", s);
    if (errno == ERANGE)
        die(false, "Number is out of range: %s", s);

    return r;
}

int
main(unsigned argc, char * argv[])
{
    struct utsname un;
    uname(&un);
    if (!strncmp(un.release, "2.4", 3))
    {
        MIN_SPEED_FILE = PROC_MIN_SPEED_FILE;
        MAX_SPEED_FILE = PROC_MAX_SPEED_FILE;
        CURRENT_SPEED_FILE = PROC_CURRENT_SPEED_FILE;
        GOVERNOR_FILE = PROC_GOVERNOR_FILE;
        USERSPACE = PROC_USERSPACE;
    }
    else
    {
        MIN_SPEED_FILE = SYSFS_MIN_SPEED_FILE;
        MAX_SPEED_FILE = SYSFS_MAX_SPEED_FILE;
        CURRENT_SPEED_FILE = SYSFS_CURRENT_SPEED_FILE;
        GOVERNOR_FILE = SYSFS_GOVERNOR_FILE;
        USERSPACE = SYSFS_USERSPACE;
    }

    unsigned interval = 20; // 2 seconds
    bool daemonize = false;

    // parse argv
    for(unsigned i = 1; i < argc; i++)
    {
        if(!strcmp(argv[i], "-d"))
            daemonize = true;
        else if(!strcmp(argv[i], "-i"))
        {
            if (argc <= i + 1)
                die(false, "The -i option must be followed by an interval in tenths of a second\n");

            interval = get_int(argv[++i]);
            errprintf("Interval is %u\n", interval);

        }
        else if(!strcmp(argv[i], "-p"))
        {
            if (argc <= i + 3)
                die(false, "The -p option must be followed by 2 integers\n");

            clock_up_idle_fast = get_int(argv[++i]);
            idle_threshold = get_int(argv[++i]);
            errprintf("Triggers are %u %u\n",clock_up_idle_fast, idle_threshold);
        }
        else if (!strcmp(argv[i], "-t"))
        {
            if (argc <= i + 2)
                die(false, "The -t option must be followed by a filename and a temperature\n");

            temperature_filename = argv[++i];
            max_temperature = get_int(argv[++i]);
        }
        else if (!strcmp(argv[i], "-a"))
        {
            if (argc <= i + 1)
                die(false, "The -a option must be followed by a filename\n");

            ac_filename = argv[++i];
        }
        else
        {
            fprintf
            (
                stderr,
                "%s v%s\n"
                "\n"
                "This program monitors the system's idle percentage and reduces or raises the\n"
                "CPU clock speed and voltage accordingly to minimize power consumption when\n"
                "idle and maximize performance when needed.  This is the default behavior.\n"
                "\n"
                "The program can also be configured to reduce the CPU clock speed if the CPU\n"
                "temperature gets too high or if the computer's AC adapter is disconnected.\n"
                "\n"
                "Usage: %s [Options]\n"
                "\n"
                "    Options:\n"
                "        -d\n"
                "            Tells the process to daemonize itself (run in background).\n"
                "\n"
                "        -i <interval>\n"
                "            Sets the interval between idle percentage tests and possible\n"
                "            speed changes in 1/10 second increments (default is 20).\n"
                "\n"
                "        -p <fast up> <threshold>\n"
                "            Sets the CPU idle percentage thresholds.  <fast up> is the idle\n"
                "            percentage below which the CPU will be set to the highest possible\n"
                "            speed.  <threshold> is the idle percentage above which the CPU\n"
                "            speed will be decreased and below which the CPU speed will be\n"
                "            increased (defaults are 10 and 25).\n"
                "\n"
                "        -t <temp file> <maxtemp>\n"
                "            Sets the ACPI temperature file and the temperature at which CPU\n"
                "            will be set to minimum speed.\n"
                "\n"
                "        -a <AC file>\n"
                "            Sets the ACPI AC adapter state file and tells the program to set\n"
                "            the CPU to minimum speed when AC adapter is disconnected.\n"
                "\n"
                "    To tell the program to stay at the highest CPU clock speed to maximize\n"
                "    performance, send the process the SIGUSR1 signal.\n"
                 "\n"
                "    To tell the program to stay at the lowest CPU clock speed to maximize\n"
                "    battery life, send the process the SIGUSR2 signal.\n"
                 "\n"
                "    To tell the program to go back to scaling the CPU clock speed dynamically,\n"
                "    send the process the SIGHUP signal.\n"
                 "\n"
                "Example:\n"
                "    cpuspeed -d -i 10 -t /proc/acpi/thermal_zone/THRM/temperature 79 \\\n"
                "        -a /proc/acpi/ac_adapter/ACAD/state\n"
                "\n"
                "    This will have the program check (and possibly change) the CPU speed once\n"
                "    every second and go to minimum CPU speed if the temperature goes above 79\n"
                "    or the AC adapter is disconnected.\n"
                "\n"
                "Author:\n"
                "    Carl Thompson <cet@carlthompson.net>\n"
                "\n"
                , NAME, VERSION, NAME
            );
            exit(0);
        }
    }

    // use the userspace governor
    write_line(GOVERNOR_FILE, "%s\n", USERSPACE);

    if (access(CURRENT_SPEED_FILE, W_OK) < 0)
        die(true, "Cannot write to speed control file: %s", CURRENT_SPEED_FILE);

    // run in background if requested
    if (daemonize)
        daemon(0, 0);

    // set up signal handling
    struct sigaction signal_action;

    // block all of our signals during each signal handler to avoid possible
    // race condition (not that it would really matter)
    sigemptyset(&signal_action.sa_mask);
    sigaddset(&signal_action.sa_mask, SIGALRM);
    sigaddset(&signal_action.sa_mask, SIGUSR1);
    sigaddset(&signal_action.sa_mask, SIGUSR2);
    sigaddset(&signal_action.sa_mask, SIGHUP);
    signal_action.sa_flags = 0;

    // set the SIGALRM handler to our function
    signal_action.sa_handler = alarm_handler;
    sigaction(SIGALRM, &signal_action, 0);

    // set the SIGUSR1 handler to our function
    signal_action.sa_handler = usr1_handler;
    sigaction(SIGUSR1, &signal_action, 0);

    // set the SIGUSR2 handler to our function
    signal_action.sa_handler = usr2_handler;
    sigaction(SIGUSR2, &signal_action, 0);

    // set the HUP handler to our function
    signal_action.sa_handler = hup_handler;
    sigaction(SIGHUP, &signal_action, 0);

    // reset the speed steps
    get_supported_speeds();
    // reset the time counters
    reset_times();

    // we dynamically scale speed to start
    mode = DYNAMIC;

    // main loop
    while (1)
    {
        raise(SIGALRM);
        usleep(interval * 100000L);
    }

    // never reached
    return 0;
}

/* $Id: cpuspeed.cc 4 2003-11-07 23:06:53Z carl $ */
